import { EventDispatcher, Quaternion, Vector3 } from "three";
import { MovementState } from "./MovementState";
import { Settings } from "../settings";
import { Updatable } from "../core";
/**
 * A translation manager.
 */
export declare class TranslationManager extends EventDispatcher implements Updatable {
    /**
     * The current position.
     */
    private position;
    /**
     * The quaternion.
     */
    private quaternion;
    /**
     * The current target.
     */
    private target;
    /**
     * The settings.
     */
    private settings;
    /**
     * The current movement state.
     */
    private movementState;
    /**
     * The current velocity.
     */
    private velocity0;
    /**
     * The target velocity.
     */
    private velocity1;
    /**
     * Scalar dampers.
     */
    private scalarDampers;
    /**
     * A timestamp.
     */
    private timestamp;
    /**
     * A reusable update event.
     */
    private updateEvent;
    /**
     * Constructs a new translation manager.
     *
     * @param position - The position.
     * @param quaternion - The quaternion.
     * @param target - The target.
     * @param settings - The settings.
     */
    constructor(position: Vector3, quaternion: Quaternion, target: Vector3, settings: Settings);
    /**
     * Returns the movement state.
     *
     * @return The movement state.
     */
    getMovementState(): MovementState;
    /**
     * Sets the position.
     *
     * @param position - A position.
     * @return This manager.
     */
    setPosition(position: Vector3): TranslationManager;
    /**
     * Sets the quaternion.
     *
     * @param quaternion - A quaternion.
     * @return This manager.
     */
    setQuaternion(quaternion: Quaternion): TranslationManager;
    /**
     * Sets the target.
     *
     * @param target - A target.
     * @return This manager.
     */
    setTarget(target: Vector3): TranslationManager;
    /**
     * Resets the current velocity.
     *
     * @return This manager.
     */
    resetVelocity(): void;
    /**
     * Translates the position along a given axis.
     *
     * @param axis - The axis.
     * @param distance - The distance.
     */
    private translateOnAxis;
    /**
     * Changes the position based on the current velocity and elapsed time.
     *
     * @param elapsed - The time since the last frame in seconds.
     */
    private translate;
    update(timestamp: number): void;
}
