import { EventDispatcher } from "three";
import { ControlMode } from "../core/ControlMode";
/**
 * JSON representation of general settings.
 */
export interface GeneralSettingsJSON {
    mode: ControlMode;
}
/**
 * General settings.
 */
export declare class GeneralSettings extends EventDispatcher {
    /**
     * The control mode.
     */
    private mode;
    /**
     * The previous control mode.
     */
    private previousMode;
    /**
     * Constructs new general settings.
     */
    constructor();
    /**
     * Returns the previous control mode.
     *
     * @return The mode.
     */
    getPreviousMode(): ControlMode;
    /**
     * Returns the control mode.
     *
     * @return The mode.
     */
    getMode(): ControlMode;
    /**
     * Sets the control mode.
     *
     * @param value - The value.
     */
    setMode(value: ControlMode): void;
    /**
     * Copies the given general settings.
     *
     * @param settings - General settings.
     * @return This instance.
     */
    copy(settings: GeneralSettings): GeneralSettings;
    /**
     * Clones this general settings instance.
     *
     * @return The cloned general settings.
     */
    clone(): GeneralSettings;
    /**
     * Copies the given JSON data.
     *
     * @param json - The JSON data.
     * @return This instance.
     */
    fromJSON(json: GeneralSettingsJSON): GeneralSettings;
    toJSON(): Record<string, unknown>;
}
