import { EventDispatcher } from "three";
import { PointerBehaviour } from "../input";
/**
 * JSON representation of pointer settings.
 */
export interface PointerSettingsJSON {
    behaviour: PointerBehaviour;
    sensitivity: number;
}
/**
 * Pointer settings.
 */
export declare class PointerSettings extends EventDispatcher {
    /**
     * The pointer behaviour.
     */
    private behaviour;
    /**
     * The baseline pointer sensitivity.
     */
    private sensitivity;
    /**
     * Constructs new pointer settings.
     */
    constructor();
    /**
     * Returns the pointer behaviour.
     *
     * @return The pointer behaviour.
     */
    getBehaviour(): PointerBehaviour;
    /**
     * Sets the pointer behaviour.
     *
     * @param value - The behaviour.
     */
    setBehaviour(value: PointerBehaviour): void;
    /**
     * Returns the sensitivity. The default is 1e-3.
     *
     * @return The sensitivity.
     */
    getSensitivity(): number;
    /**
     * Sets the sensitivity.
     *
     * This sensitivity acts as a baseline scale for pointer movement deltas.
     *
     * @param value - The sensitivity.
     */
    setSensitivity(value: number): void;
    /**
     * Copies the given pointer settings.
     *
     * @param settings - Pointer settings.
     * @return This instance.
     */
    copy(settings: PointerSettings): PointerSettings;
    /**
     * Clones this pointer settings instance.
     *
     * @return The cloned pointer settings.
     */
    clone(): PointerSettings;
    /**
     * Copies the given JSON data.
     *
     * @param json - The JSON data.
     * @return This instance.
     */
    fromJSON(json: PointerSettingsJSON): PointerSettings;
    toJSON(): Record<string, unknown>;
}
